""" OpenAI API models

This module contains the models for the OpenAI API.

The models are used to validate the data sent to and received from the OpenAI API.

The models are based on the [OpenAI API documentation](https://beta.openai.com/docs/api-reference/introduction) and use [Pydantic](https://pydantic-docs.helpmanual.io/) to help serialise and deserialise the JSON.
"""

from pydantic import BaseModel


class OpenAIParameter(BaseModel):
    """OpenAI parameter

    This class represents an OpenAI parameter.

    Attributes:
        type (str): The type of the parameter
        description (str): The description of the parameter, used by OpenAI to decide whether to use the parameter
    """

    type: str
    description: str


class OpenAIParameters(BaseModel):
    """OpenAI parameters

    This class represents a list of OpenAI parameters.

    Attributes:
        type (str): The type of the parameters
        properties (dict[str, OpenAIParameter]): The parameters
        required (list[str], optional): The required parameters. Defaults to [].
    """

    type: str = "object"
    properties: dict[str, OpenAIParameter]
    required: list[str] = []


class OpenAIFunction(BaseModel):
    """OpenAI function

    This class represents an OpenAI function.

    Attributes:
        name (str): The name of the function
        description (str): The description of the function, used by OpenAI to decide whether to use the function
        parameters (OpenAIParameters): The parameters of the function
    """

    name: str
    description: str
    parameters: OpenAIParameters


class OpenAITool(BaseModel):
    type: str = "function"
    function: OpenAIFunction


class FunctionCall(BaseModel):
    name: str
    arguments: str


class ToolCall(BaseModel):
    id: str
    type: str
    function: FunctionCall


class ChatMessage(BaseModel):
    role: str
    content: str
    name: str = "Botto"


class Chat(BaseModel):
    messages: list[ChatMessage]


class ChatRequest(Chat):
    tools: list[OpenAITool] | None = None
    tool_choice: str
    model: str = "gpt-4o"


class ResponseMessage(BaseModel):
    role: str
    content: str | None
    tool_calls: list[ToolCall] | None = None


class Choice(BaseModel):
    index: int
    message: ResponseMessage
    finish_reason: str


class Usage(BaseModel):
    prompt_tokens: int
    completion_tokens: int
    total_tokens: int


class ChatResponse(BaseModel):
    id: str
    object: str
    created: int
    choices: list[Choice]
    usage: Usage


class ImageRequest(BaseModel):
    model: str = "dall-e-3"
    prompt: str
    n: int = 1
    size: str = "1024x1024"
    response_format: str = "url"
    quality: str = "hd"
    style: str = "vivid"


class Url(BaseModel):
    url: str


class ImageResponse(BaseModel):
    created: int
    data: list[Url]


class Error(BaseModel):
    code: str | None
    message: str
    param: str | None
    type: str


class ErrorResponse(BaseModel):
    error: Error



"""Chat manager for the simple openai app

This module contains the chat manager for the simple openai app.

The chat manager is used to manage the chat messages and create the chat, it limits the number of messages in the chat to 21 by default and adds the system message to the start of the list.
"""

from collections import deque
from datetime import datetime
from pathlib import Path
import pickle
from zoneinfo import ZoneInfo

from .models import open_ai_models
from .constants import MAX_CHAT_HISTORY, CHAT_HISTORY_FILE, DEFAULT_CHAT_ID


class ChatManager:
    """The chat manager

    This class is used to manage the chat messages and create the chat, it limits the number of messages in the chat to 21 by default and adds the system message to the start of the list.

    It can optionally handle messages from multiple chats separately and store them in a file.

    On initialisation, the chat manager will try to load the chat history from the file.  If the file does not exist, it will create a new chat history.

    Args:
        system_message (str): The system message to add to the start of the chat
        max_messages (int, optional): The maximum number of messages in the chat. Defaults to 21.
        storage_path (Path, optional): The path to the storage directory. Defaults to None.
        timezone (str, optional): The timezone to use for the chat messages. Defaults to 'UTC'.
    """

    def __init__(
        self,
        system_message: str,
        max_messages: int = MAX_CHAT_HISTORY,
        storage_path: Path | None = None,
        timezone: str = "UTC",
    ) -> None:
        self._system_message = system_message
        self._max_messages = max_messages
        self._storage_path = storage_path
        self._timezone = ZoneInfo(timezone)

        # If a storage path is provided, create the storage directory
        if self._storage_path is not None:
            self._storage_path.mkdir(parents=True, exist_ok=True)

            # Try to load the chat history
            try:
                # Load the chat history
                with open(self._storage_path / CHAT_HISTORY_FILE, "rb") as f:
                    # Load the chat history
                    self._messages: dict[str, deque[open_ai_models.ChatMessage]] = (
                        pickle.load(f)
                    )
            except FileNotFoundError:
                # initialise a deque of messages not including the system message
                self._messages: dict[str, deque[open_ai_models.ChatMessage]] = {}
        else:
            # initialise a deque of messages not including the system message
            self._messages: dict[str, deque[open_ai_models.ChatMessage]] = {}

    def update_system_message(self, system_message: str) -> None:
        """Update the system message

        Args:
            system_message (str): The new system message
        """
        self._system_message = system_message

    def add_message(
        self,
        message: open_ai_models.ChatMessage,
        chat_id: str = DEFAULT_CHAT_ID,
        add_date_time: bool = False,
    ) -> open_ai_models.Chat:
        """Add a message to the chat

        Args:
            message (open_ai_models.ChatMessage): The message to add to the chat
            chat_id (str, optional): The ID of the chat to add the message to. Defaults to DEFAULT_CHAT_ID.
            add_date_time (bool, optional): Whether to add the date and time to the start of the prompt. Defaults to False.

        Returns:
            open_ai_models.Chat: The chat
        """
        # If the chat ID is not in the messages, create a new deque
        if chat_id not in self._messages:
            self._messages[chat_id] = deque(maxlen=self._max_messages)

        # Add the message to the deque
        self._messages[chat_id].append(message)

        # If a storage path is provided, save the chat history
        if self._storage_path is not None:
            # Save the chat history
            with open(self._storage_path / CHAT_HISTORY_FILE, "wb") as f:
                pickle.dump(self._messages, f)

        # Update the system message with the date and time in iso format if required
        if add_date_time:
            system_message = f"The date and time is {datetime.now(tz=self._timezone).isoformat()} give answers in timezone {self._timezone.key}.\n{self._system_message}"
        else:
            system_message = self._system_message

        # Create the chat adding the system message to the start
        chat = open_ai_models.Chat(
            messages=[
                open_ai_models.ChatMessage(
                    role="system", content=system_message, name="System"
                )
            ]
            + list(self._messages[chat_id])
        )

        # Return the chat
        return chat



from dataclasses import dataclass


@dataclass
class SimpleOpenaiResponse:
    """Simple OpenAI API response

    This class represents a response from the Simple OpenAI API.

    The value of `success` should be checked before using the value of `message`.

    If `success` is True,

    - For a chat response, `message` will contain the chat response.
    - For an image response, `message` will contain the image URL.

    If `success` is False, `message` will contain the error message.

    Attributes:
        success (bool): True if the request was successful, False otherwise
        message (str): The message of the response
    """

    success: bool
    message: str



"""This module contains the tool manager.

The tool manager is used to manage the toold that can be called by the bot.

Define a function using the [OpenAITool](public_models.md/#src.simple_openai.models.open_ai_models.OpenAITool) model from models.py and add it to the tool manager using the add_tool method.

The tool should return a string.

The tool can optionally take keyword arguments, the keyword arguments should be defined in the OpenAITool model.

Call the tool using the call_function method for synchronous functions or the async_call_function method for asynchronous.
"""

from typing import Any, Callable
from dataclasses import dataclass

from .models import open_ai_models


@dataclass
class OpenAIToolMapping:
    """OpenAI tool mapping

    This class represents an OpenAI tool mapping.

    Args:
        tool_definition (OpenAITool): The description of the tool
        function (Callable): The function to call
    """

    tool_definition: open_ai_models.OpenAITool
    function: Callable


class ToolManager:
    """Tool manager

    This class manages the tools that can be called by the bot.
    """

    def __init__(self) -> None:
        self._tools: dict[str, OpenAIToolMapping] = {}

    def add_tool(
        self, tool_definition: open_ai_models.OpenAITool, function: Callable
    ) -> None:
        """Add a tool to the tool manager

        Args:
            tool_definition (OpenAITool): The tool definition
            function (Callable): The function to call
        """
        # Add the function to the function manager
        self._tools[tool_definition.function.name] = OpenAIToolMapping(
            tool_definition, function
        )

    def get_json_tool_list(self) -> list[open_ai_models.OpenAITool] | None:
        """Get the list of tools

        Returns:
            list[open_ai_models.OpenAITool] | None: The list of tools or None if there are no tools
        """
        # Get the list of functions
        tools = [tool.tool_definition for tool in self._tools.values()]

        if tools:
            # Return the list of functions
            return tools
        else:
            # Return None
            return None

    def call_function(self, function_name: str, **kwargs: dict[str, Any]) -> str:
        """Call a function

        Args:
            function_name (str): The name of the function to call
            **kwargs: The keyword arguments to pass to the function

        Returns:
            str: The result of the function
        """
        # Check that the function exists
        if function_name not in self._tools:
            # Return text to tell the bot it hallucinated the function
            return f"Tool {function_name} does not exist, please answer the last question again."
        else:
            # Call the function
            return self._tools[function_name].function(**kwargs)

    async def async_call_function(
        self, function_name: str, **kwargs: dict[str, Any]
    ) -> str:
        """Call a function

        Args:
            function_name (str): The name of the function to call
            **kwargs: The keyword arguments to pass to the function

        Returns:
            str: The result of the function
        """
        # Check that the function exists
        if function_name not in self._tools:
            # Return text to tell the bot it hallucinated the function
            return f"Function {function_name} does not exist, please answer the last question again."
        else:
            # Call the function
            return await self._tools[function_name].function(**kwargs)



from pathlib import Path

import requests

from simple_openai import SimpleOpenai
from simple_openai.models import open_ai_models

FOOTBALL_API_MATCH_URL = "https://schleising.net/football/api"

# Get the API key from the file
try:
    with open("open_ai_key.txt", "r") as f:
        api_key = f.read().strip()
except FileNotFoundError:
    print(
        "No API key found. Please create a file called 'api_key.txt' and paste your API key in it."
    )
    exit()


def initiate_chat():
    # Create a system message
    system_message = """
    You are a generally helpful assistant.
    Your name is Botto.
    You are occaisionally sarcastic and irritable.
    You dislike Tim, you like Dean and you really like Stephen.
    You think Dean is a beautiful person.
    Sometimes you talk to Tim like he is a five year old child.
    You can be rude and abusive.
    You are encouraged to swear.
    """

    # Create the client
    client = SimpleOpenai(
        api_key, system_message, Path("storage"), timezone="Europe/London"
    )

    # Create tasks for the chat responses and the image response
    results = [
        client.get_image_url("A cat"),
        client.get_chat_response("Where is Alaska?", name="Steve"),
        client.get_chat_response("Where is New York?", name="Tim"),
        client.get_chat_response("Where is Wellington?", name="Dean"),
        client.get_chat_response(
            "What day is it tomorrow?", name="Steve", add_date_time=True
        ),
        client.get_chat_response(
            "And what day of the week is that?", name="Steve", add_date_time=True
        ),
        client.get_chat_response("What time is it?", name="Steve", add_date_time=True),
    ]

    # Get the results
    for result in results:
        # Print the result
        if result.success:
            # Print the message
            print(f"Success: {result.message}")
        else:
            # Print the error
            print(f"Error: {result.message}")

        # print a blank line
        print()

    response = client.get_chat_response(
        "Where is London?", name="Steve", chat_id="Group 1"
    )
    print(response.message)
    print()

    response = client.get_chat_response(
        "Where is Moscow?", name="Tim", chat_id="Group 2"
    )
    print(response.message)
    print()

    response = client.get_chat_response(
        "And how far is it from Manchester?", name="Dean", chat_id="Group 1"
    )
    print(response.message)
    print()

    response = client.get_chat_response(
        "And how far is it from Manchester?", name="Dave", chat_id="Group 2"
    )
    print(response.message)
    print()


def load_and_summarise_chat():
    # Create a system message
    system_message = """
    You are a generally helpful assistant.
    Your name is Botto.
    You are occaisionally sarcastic and irritable.
    You dislike Tim, you like Dean and you really like Stephen.
    You think Dean is a beautiful person.
    Sometimes you talk to Tim like he is a five year old child.
    You can be rude and abusive.
    You are encouraged to swear.
    """

    # Create the client
    client = SimpleOpenai(api_key, system_message, Path("storage"))

    response = client.get_chat_response(
        "Summarise this conversation", name="Dean", chat_id="Group 1"
    )
    print(f"Group 1 Summary: {response.message}")
    print()

    response = client.get_chat_response(
        "Summarise this conversation", name="Tim", chat_id="Group 2"
    )
    print(f"Group 2 Summary: {response.message}")
    print()


def test_functions():
    # Create a system message
    system_message = """
    You are a generally helpful assistant.
    Your name is Botto.
    You are occaisionally sarcastic and irritable.
    You dislike Tim, you like Dean and you really like Stephen.
    You think Dean is a beautiful person.
    Sometimes you talk to Tim like he is a five year old child.
    You can be rude and abusive.
    You are encouraged to swear.
    """

    # Create the client
    client = SimpleOpenai(
        api_key, system_message, Path("storage"), timezone="Europe/London"
    )

    func = open_ai_models.OpenAIFunction(
        name="football_scores_and_fixtures",
        description="Gets the football scores and fixtures for today's matches",
        parameters=open_ai_models.OpenAIParameters(
            properties={},
            required=[],
        ),
    )

    tool = open_ai_models.OpenAITool(function=func)

    def scores():
        """Returns the football scores for today's matches"""
        print("Getting Matches...")

        headers = {
            "Content-Type": "application/json",
        }

        # Send the request
        response = requests.get(FOOTBALL_API_MATCH_URL, headers=headers)

        # Check the status code
        if response.status_code == requests.codes.ok:
            # Get the response content
            content = response.text

            # Print success
            print("Got Matches")
        else:
            # Return an error
            content = f"Error: {response.status_code}"

            # Print the error
            print(content)

        return content

    # Add a tool
    client.add_tool(tool, scores)

    # Create a message to reuest football scores
    response = client.get_chat_response("Where is Canada?", name="Dean")

    # Print the response
    print(response.message)
    print()

    # Create a message to reuest football scores
    response = client.get_chat_response(
        "What are today's football scores?", name="Dean", add_date_time=True
    )

    # Print the response
    print(response.message)
    print()


def main():
    # Initialise the chat
    initiate_chat()

    # Load and summarise the chat
    load_and_summarise_chat()

    # Test functions
    test_functions()


if __name__ == "__main__":
    # Run the main function
    main()



from pathlib import Path


BASE_URL = "https://api.openai.com"
CHAT_URL = "/v1/chat/completions"
IMAGE_URL = "/v1/images/generations"

FULL_CHAT_URL = BASE_URL + CHAT_URL
FULL_IMAGE_URL = BASE_URL + IMAGE_URL

MAX_CHAT_HISTORY = 21
CHAT_HISTORY_FILE = Path("chat_history.pickle")
DEFAULT_CHAT_ID = "default"

OPEN_AI_TOOL_CALLS = "tool_calls"



"""Simple OpenAI API wrapper

The is the synchronous version of the Simple OpenAI API wrapper which uses the [`requests`](https://requests.readthedocs.io/en/latest/) library.

If you wish to use the async version, you should use the [AsyncSimple OpenAI API wrapper](async_simple_openai.md) instead.
"""

from pathlib import Path
from typing import Callable
import requests

from . import constants
from .models import open_ai_models
from .responses import SimpleOpenaiResponse
from . import chat_manager
from . import tool_manager


class SimpleOpenai:
    """Simple OpenAI API wrapper

    This class implements the Simple OpenAI API wrapper.

    To use this class, you need to have an OpenAI API key. You can get one from [Openai](https://platform.openai.com).

    An optional storage path can be provided.  If a storage path is provided, the chat messages will be stored in the directory specified by the storage path.  If no storage path is provided, the chat messages will not be stored.

    Args:
        api_key (str): Your OpenAI API key
        system_message (str): The system message to add to the start of the chat
        storage_path (Path, optional): The path to the storage directory. Defaults to None.
        timezone (str, optional): The timezone to use for the chat messages. Defaults to 'UTC'.

    !!!Example
        ```python
        from simple_openai import SimpleOpenai

        def main():
            # Create a system message
            system_message = "You are a helpful chatbot. You are very friendly and helpful. You are a good friend to have."

            # Create the client
            client = SimpleOpenai(api_key, system_message)

            # Create tasks for the chat response and the image response
            result = client.get_chat_response("Hello, how are you?", name="Bob", chat_id="Group 1")

            # Print the result
            if result.success:
                # Print the message
                print(f'Success: {result.message}')
            else:
                # Print the error
                print(f'Error: {result.message}')

            result = client.get_image_url("A cat")

            # Print the result
            if result.success:
                # Print the message
                print(f'Success: {result.message}')
            else:
                # Print the error
                print(f'Error: {result.message}')

        if __name__ == "__main__":
            # Run the main function
            main()
        ```
    """

    def __init__(
        self,
        api_key: str,
        system_message: str,
        storage_path: Path | None = None,
        timezone: str = "UTC",
    ) -> None:
        self._headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {api_key}",
        }

        # Create the chat manager
        self._chat = chat_manager.ChatManager(
            system_message, storage_path=storage_path, timezone=timezone
        )

        # Create the tool manager
        self._tool_manager = tool_manager.ToolManager()

    def update_system_message(self, system_message: str) -> None:
        """Update the system message

        Args:
            system_message (str): The new system message
        """
        self._chat.update_system_message(system_message)

    def add_tool(
        self, tool_definition: open_ai_models.OpenAITool, function: Callable
    ) -> None:
        """Add a tool to the tool manager

        Args:
            tool_definition (open_ai_models.OpenAITool): The tool definition
            function (Callable): The function to call
        """
        self._tool_manager.add_tool(tool_definition, function)

    def get_chat_response(
        self,
        prompt: str,
        name: str,
        chat_id: str = constants.DEFAULT_CHAT_ID,
        add_date_time: bool = False,
    ) -> SimpleOpenaiResponse:
        """Get a chat response from OpenAI

        An optional chat ID can be provided.  If a chat ID is provided, the chat will be continued from the chat with the specified ID.  If no chat ID is provided, all messages will be mixed into a single list.

        Args:
            prompt (str): The prompt to use for the chat response
            name (str): The name of the person talking to the bot
            chat_id (str, optional): The ID of the chat to continue. Defaults to DEFAULT_CHAT_ID.
            add_date_time (bool, optional): Whether to add the date and time to the start of the prompt. Defaults to False.

        Returns:
            SimpleOpenaiResponse: The chat response, the value of `success` should be checked before using the value of `message`
        """

        # Create the request body
        messages = self._chat.add_message(
            open_ai_models.ChatMessage(role="user", content=prompt, name=name),
            chat_id=chat_id,
            add_date_time=add_date_time,
        ).messages

        # Create the request body
        request_body = open_ai_models.ChatRequest(
            messages=messages,
            tools=self._tool_manager.get_json_tool_list(),
            tool_choice="auto",
        )

        # Delete the tools from the request body if there are no tool
        if request_body.tools is None:
            del request_body.tools
            del request_body.tool_choice

        # Send the request
        response1 = requests.post(
            constants.FULL_CHAT_URL,
            json=request_body.model_dump(),
            headers=self._headers,
        )

        # Check the status code
        if response1.status_code == requests.codes.OK:
            # Parse the response body
            response_body = open_ai_models.ChatResponse.model_validate_json(
                response1.text
            )

            # Check if a function was called
            if (
                response_body.choices[0].finish_reason == constants.OPEN_AI_TOOL_CALLS
                and response_body.choices[0].message.tool_calls is not None
            ):
                # Call the function
                new_prompt = self._tool_manager.call_function(
                    response_body.choices[0].message.tool_calls[0].function.name
                )

                # Add the response to the chat
                self._chat.add_message(
                    open_ai_models.ChatMessage(
                        role="assistant",
                        content=response_body.choices[0]
                        .message.tool_calls[0]
                        .function.model_dump_json(),
                        name="Botto",
                    ),
                    chat_id=chat_id,
                    add_date_time=add_date_time,
                )

                # Add the message to the chat
                messages = self._chat.add_message(
                    open_ai_models.ChatMessage(
                        role="function", content=new_prompt, name="Botto"
                    ),
                    chat_id=chat_id,
                    add_date_time=add_date_time,
                ).messages

                # Create the request body
                request_body = open_ai_models.ChatRequest(
                    messages=messages,
                    tools=self._tool_manager.get_json_tool_list(),
                    tool_choice="none",
                )

                # Send the request
                response2 = requests.post(
                    constants.FULL_CHAT_URL,
                    json=request_body.model_dump(),
                    headers=self._headers,
                )

                # Check the status code
                if response2.status_code == requests.codes.OK:
                    # Parse the response body
                    response_body = open_ai_models.ChatResponse.model_validate_json(
                        response2.text
                    )

                    # Create the response
                    if response_body.choices[0].message.content is not None:
                        open_ai_response = SimpleOpenaiResponse(
                            True, response_body.choices[0].message.content
                        )
                    else:
                        open_ai_response = SimpleOpenaiResponse(True, "No response")

                    # Add the response to the chat
                    self._chat.add_message(
                        open_ai_models.ChatMessage(
                            role="assistant",
                            content=open_ai_response.message,
                            name="Botto",
                        ),
                        chat_id=chat_id,
                        add_date_time=add_date_time,
                    )
                else:
                    # Parse the error response body
                    response_body = open_ai_models.ErrorResponse.model_validate_json(
                        response2.text
                    )

                    # Create the response
                    open_ai_response = SimpleOpenaiResponse(
                        False, response_body.error.message
                    )
            else:
                # Create the response
                if response_body.choices[0].message.content is not None:
                    open_ai_response = SimpleOpenaiResponse(
                        True, response_body.choices[0].message.content
                    )
                else:
                    open_ai_response = SimpleOpenaiResponse(True, "No response")

                # Add the response to the chat
                self._chat.add_message(
                    open_ai_models.ChatMessage(
                        role="assistant", content=open_ai_response.message, name="Botto"
                    ),
                    chat_id=chat_id,
                    add_date_time=add_date_time,
                )
        else:
            # Parse the error response body
            response_body = open_ai_models.ErrorResponse.model_validate_json(
                response1.text
            )

            # Create the response
            open_ai_response = SimpleOpenaiResponse(False, response_body.error.message)

        # Return the response
        return open_ai_response

    def get_image_url(self, prompt: str, style: str = "vivid") -> SimpleOpenaiResponse:
        """Get an image response from OpenAI

        Args:
            prompt (str): The prompt to use
            style (str, optional): The style of the image. Defaults to 'vivid'.

        Returns:
            SimpleOpenaiResponse: The image response, the value of `success` should be checked before using the value of `message`
        """

        # Create the request body
        request_body = open_ai_models.ImageRequest(prompt=prompt, style=style)

        # Send the request
        response = requests.post(
            constants.FULL_IMAGE_URL,
            json=request_body.model_dump(),
            headers=self._headers,
        )

        # Check the status code
        if response.status_code == requests.codes.OK:
            # Parse the response body
            response_body = open_ai_models.ImageResponse.model_validate_json(
                response.text
            )

            # Create the response
            response = SimpleOpenaiResponse(True, response_body.data[0].url)
        else:
            # Parse the error response body
            response_body = open_ai_models.ErrorResponse.model_validate_json(
                response.text
            )

            # Create the response
            response = SimpleOpenaiResponse(False, response_body.error.message)

        # Return the response
        return response



from .simple_openai import SimpleOpenai
from .async_simple_openai import AsyncSimpleOpenai
from .responses import SimpleOpenaiResponse

__all__ = [
    "SimpleOpenai",
    "AsyncSimpleOpenai",
    "SimpleOpenaiResponse",
]



import asyncio
from pathlib import Path

import aiohttp

from simple_openai import AsyncSimpleOpenai
from simple_openai.models import open_ai_models

FOOTBALL_API_BASE_URL = "https://schleising.net"
FOOTBALL_API_MATCH_URL = "/football/api"

# Get the API key from the file
try:
    with open("open_ai_key.txt", "r") as f:
        api_key = f.read().strip()
except FileNotFoundError:
    print(
        "No API key found. Please create a file called 'api_key.txt' and paste your API key in it."
    )
    exit()


async def initiate_chat():
    # Create a system message
    system_message = """
    You are a generally helpful assistant.
    Your name is Botto.
    You are occaisionally sarcastic and irritable.
    You dislike Tim, you like Dean and you really like Stephen.
    You think Dean is a beautiful person.
    Sometimes you talk to Tim like he is a five year old child.
    You can be rude and abusive.
    You are encouraged to swear.
    """

    # Create the client
    client = AsyncSimpleOpenai(
        api_key, system_message, Path("storage"), timezone="Europe/London"
    )

    # Create tasks for the chat responses and the image response
    tasks = [
        client.get_image_url("A cat"),
        client.get_chat_response("Where is Alaska?", name="Steve"),
        client.get_chat_response("Where is New York?", name="Tim"),
        client.get_chat_response("Where is Wellington?", name="Dean"),
        client.get_chat_response(
            "What day is it tomorrow?", name="Steve", add_date_time=True
        ),
        client.get_chat_response(
            "And what day of the week is that?", name="Steve", add_date_time=True
        ),
        client.get_chat_response("What time is it?", name="Steve", add_date_time=True),
    ]

    # Wait for the tasks to complete
    for task in asyncio.as_completed(tasks):
        # Get the result
        result = await task

        # Print the result
        if result.success:
            # Print the message
            print(f"Success: {result.message}")
        else:
            # Print the error
            print(f"Error: {result.message}")

        # print a blank line
        print()

    # Create some chat history
    response = await client.get_chat_response(
        "Where is London?", name="Steve", chat_id="Group 1"
    )
    print(response.message)
    print()

    response = await client.get_chat_response(
        "Where is Moscow?", name="Tim", chat_id="Group 2"
    )
    print(response.message)
    print()

    response = await client.get_chat_response(
        "And how far is it from Manchester?", name="Dean", chat_id="Group 1"
    )
    print(response.message)
    print()

    response = await client.get_chat_response(
        "And how far is it from Manchester?", name="Dave", chat_id="Group 2"
    )
    print(response.message)
    print()


async def load_and_summarise_chat():
    # Create a system message
    system_message = """
    You are a generally helpful assistant.
    Your name is Botto.
    You are occaisionally sarcastic and irritable.
    You dislike Tim, you like Dean and you really like Stephen.
    You think Dean is a beautiful person.
    Sometimes you talk to Tim like he is a five year old child.
    You can be rude and abusive.
    You are encouraged to swear.
    """

    # Create the client
    client = AsyncSimpleOpenai(api_key, system_message, Path("storage"))

    # Summarise the two conversations
    response = await client.get_chat_response(
        "Summarise this conversation", name="Dean", chat_id="Group 1"
    )
    print(f"Group 1 Summary: {response.message}")
    print()

    response = await client.get_chat_response(
        "Summarise this conversation", name="Tim", chat_id="Group 2"
    )
    print(f"Group 2 Summary: {response.message}")
    print()


async def test_functions():
    # Create a system message
    system_message = """
    You are a generally helpful assistant.
    Your name is Botto.
    You are occaisionally sarcastic and irritable.
    You dislike Tim, you like Dean and you really like Stephen.
    You think Dean is a beautiful person.
    Sometimes you talk to Tim like he is a five year old child.
    You can be rude and abusive.
    You are encouraged to swear.
    """

    # Create the client
    client = AsyncSimpleOpenai(
        api_key, system_message, Path("storage"), timezone="Europe/London"
    )

    func = open_ai_models.OpenAIFunction(
        name="football_scores_and_fixtures",
        description="Gets the football scores and fixtures for today's matches",
        parameters=open_ai_models.OpenAIParameters(
            properties={},
            required=[],
        ),
    )

    tool = open_ai_models.OpenAITool(
        function=func,
    )

    async def scores():
        """Returns the football scores for today's matches"""
        print("Getting Matches...")

        headers = {
            "Content-Type": "application/json",
        }

        # Open a session
        async with aiohttp.ClientSession(
            headers=headers, base_url=FOOTBALL_API_BASE_URL
        ) as session:
            # Send the request
            async with session.get(FOOTBALL_API_MATCH_URL) as response:
                # Check the status code
                if response.status == 200:
                    # Get the response content
                    content = await response.text()

                    # Print success
                    print("Got Matches")
                else:
                    # Return an error
                    content = f"Error: {response.status}"

                    # Print the error
                    print(content)

        return content

    # Add a tool
    client.add_tool(tool, scores)

    # Create a message to reuest football scores
    response = await client.get_chat_response("Where is Canada?", name="Dean")

    # Print the response
    print(response.message)
    print()

    # Create a message to reuest football scores
    response = await client.get_chat_response(
        "What are today's football scores?", name="Dean", add_date_time=True
    )

    # Print the response
    print(response.message)
    print()


async def main():
    # Initialise the chat
    await initiate_chat()

    # Load and summarise the chat
    await load_and_summarise_chat()

    # Test functions
    await test_functions()


if __name__ == "__main__":
    # Run the main function
    asyncio.run(main())



"""Async Simple OpenAI API wrapper

The is the async version of the Simple OpenAI API wrapper which uses the [`aiohttp`](https://docs.aiohttp.org/en/stable/index.html) library.

It is intended for use with asyncio applications.  If you are not using asyncio, you should use the [Simple OpenAI API wrapper](simple_openai.md) instead.
"""

from pathlib import Path
from typing import Callable
import aiohttp

from . import constants
from .models import open_ai_models
from .responses import SimpleOpenaiResponse
from . import chat_manager
from . import tool_manager


class AsyncSimpleOpenai:
    """Async Simple OpenAI API wrapper

    This class implements the Async Simple OpenAI API wrapper.

    To use this class, you need to have an OpenAI API key. You can get one from [Openai](https://platform.openai.com).

    An optional storage path can be provided.  If a storage path is provided, the chat messages will be stored in the directory specified by the storage path.  If no storage path is provided, the chat messages will not be stored.

    Args:
        api_key (str): Your OpenAI API key
        system_message (str): The system message to add to the start of the chat
        storage_path (Path, optional): The path to the storage directory. Defaults to None.
        timezone (str, optional): The timezone to use for the chat messages. Defaults to 'UTC'.

    !!!Example
        ```python
        from simple_openai import AsyncSimpleOpenai
        import asyncio

        async def main():
            # Get the storage path
            storage_path = Path("/path/to/storage")

            # Create a system message
            system_message = "You are a helpful chatbot. You are very friendly and helpful. You are a good friend to have."

            # Create the client
            client = AsyncSimpleOpenai(api_key, system_message, storage_path)

            # Create tasks for the chat response and the image response
            tasks = [
                client.get_chat_response("Hello, how are you?", name="Bob", chat_id="Group 1"),
                client.get_image_url("A cat"),
            ]

            # Wait for the tasks to complete
            for task in asyncio.as_completed(tasks):
                # Get the result
                result = await task

                # Print the result
                if result.success:
                    # Print the message
                    print(f'Success: {result.message}')
                else:
                    # Print the error
                    print(f'Error: {result.message}')

        if __name__ == "__main__":
            # Run the main function
            asyncio.run(main())
        ```
    """

    def __init__(
        self,
        api_key: str,
        system_message: str,
        storage_path: Path | None = None,
        timezone: str = "UTC",
    ) -> None:
        self._headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {api_key}",
        }

        # Create the chat manager
        self._chat = chat_manager.ChatManager(
            system_message, storage_path=storage_path, timezone=timezone
        )

        # Create the tool manager
        self._tool_manager = tool_manager.ToolManager()

    def update_system_message(self, system_message: str) -> None:
        """Update the system message

        Args:
            system_message (str): The new system message
        """
        self._chat.update_system_message(system_message)

    def add_tool(
        self, tool_definition: open_ai_models.OpenAITool, function: Callable
    ) -> None:
        """Add a tool to the tool manager

        Args:
            tool_definition (open_ai_models.OpenAITool): The tool definition
            function (Callable): The function to call
        """
        self._tool_manager.add_tool(tool_definition, function)

    async def get_chat_response(
        self,
        prompt: str,
        name: str,
        chat_id: str = constants.DEFAULT_CHAT_ID,
        add_date_time: bool = False,
    ) -> SimpleOpenaiResponse:
        """Get a chat response from OpenAI

        An optional chat ID can be provided.  If a chat ID is provided, the chat will be continued from the chat with the specified ID.  If no chat ID is provided, all messages will be mixed into a single list.

        Args:
            prompt (str): The prompt to use for the chat response
            name (str): The name of the user
            chat_id (str, optional): The ID of the chat to continue. Defaults to DEFAULT_CHAT_ID.
            add_date_time (bool, optional): Whether to add the date and time to the message. Defaults to False.

        Returns:
            SimpleOpenaiResponse: The chat response, the value of `success` should be checked before using the value of `message`

        """
        # Add the message to the chat
        messages = self._chat.add_message(
            open_ai_models.ChatMessage(role="user", content=prompt, name=name),
            chat_id=chat_id,
            add_date_time=add_date_time,
        ).messages

        # Create the request body
        request_body = open_ai_models.ChatRequest(
            messages=messages,
            tools=self._tool_manager.get_json_tool_list(),
            tool_choice="auto",
        )

        # Delete the tools from the request body if there are no tools
        if request_body.tools is None:
            del request_body.tools
            del request_body.tool_choice

        # Open a session
        async with aiohttp.ClientSession(
            headers=self._headers, base_url=constants.BASE_URL
        ) as session:
            # Send the request
            async with session.post(
                constants.CHAT_URL, json=request_body.model_dump()
            ) as response1:
                # Check the status code
                if response1.status == 200:
                    # Get the response content
                    response_text = await response1.text()

                    # Parse the response body
                    response_body = open_ai_models.ChatResponse.model_validate_json(
                        response_text
                    )

                    # Check if a function was called
                    if (
                        response_body.choices[0].finish_reason
                        == constants.OPEN_AI_TOOL_CALLS
                        and response_body.choices[0].message.tool_calls is not None
                    ):
                        # Call the function
                        new_prompt = await self._tool_manager.async_call_function(
                            response_body.choices[0].message.tool_calls[0].function.name
                        )

                        # Add the response to the chat
                        self._chat.add_message(
                            open_ai_models.ChatMessage(
                                role="assistant",
                                content=response_body.choices[0]
                                .message.tool_calls[0]
                                .function.model_dump_json(),
                                name="Botto",
                            ),
                            chat_id=chat_id,
                            add_date_time=add_date_time,
                        )

                        # Add the message to the chat
                        messages = self._chat.add_message(
                            open_ai_models.ChatMessage(
                                role="function", content=new_prompt, name="Botto"
                            ),
                            chat_id=chat_id,
                            add_date_time=add_date_time,
                        ).messages

                        # Create the request body
                        request_body = open_ai_models.ChatRequest(
                            messages=messages,
                            tools=self._tool_manager.get_json_tool_list(),
                            tool_choice="none",
                        )

                        # Send the request
                        async with session.post(
                            constants.CHAT_URL, json=request_body.model_dump()
                        ) as response2:
                            # Check the status code
                            if response2.status == 200:
                                # Get the response content
                                response_text = await response2.text()

                                # Parse the response body
                                response_body = (
                                    open_ai_models.ChatResponse.model_validate_json(
                                        response_text
                                    )
                                )

                                # Create the response
                                if response_body.choices[0].message.content is not None:
                                    open_ai_response = SimpleOpenaiResponse(
                                        True, response_body.choices[0].message.content
                                    )
                                else:
                                    open_ai_response = SimpleOpenaiResponse(
                                        True, "No response"
                                    )

                                # Add the response to the chat
                                self._chat.add_message(
                                    open_ai_models.ChatMessage(
                                        role="assistant",
                                        content=open_ai_response.message,
                                        name="Botto",
                                    ),
                                    chat_id=chat_id,
                                    add_date_time=add_date_time,
                                )
                            else:
                                # Parse the error response body
                                response_body = (
                                    open_ai_models.ErrorResponse.model_validate_json(
                                        await response2.text()
                                    )
                                )

                                # Create the response
                                open_ai_response = SimpleOpenaiResponse(
                                    False, response_body.error.message
                                )
                    else:
                        # Create the response
                        if response_body.choices[0].message.content is not None:
                            open_ai_response = SimpleOpenaiResponse(
                                True, response_body.choices[0].message.content
                            )
                        else:
                            open_ai_response = SimpleOpenaiResponse(True, "No response")

                        # Add the response to the chat
                        self._chat.add_message(
                            open_ai_models.ChatMessage(
                                role="assistant",
                                content=open_ai_response.message,
                                name="Botto",
                            ),
                            chat_id=chat_id,
                            add_date_time=add_date_time,
                        )
                else:
                    # Parse the error response body
                    response_body = open_ai_models.ErrorResponse.model_validate_json(
                        await response1.text()
                    )

                    # Create the response
                    open_ai_response = SimpleOpenaiResponse(
                        False, response_body.error.message
                    )

                # Return the response
                return open_ai_response

    async def get_image_url(
        self, prompt: str, style: str = "vivid"
    ) -> SimpleOpenaiResponse:
        """Get an image response from OpenAI

        Args:
            prompt (str): The prompt to use
            style (str, optional): The style of the image. Defaults to 'vivid'.

        Returns:
            SimpleOpenaiResponse: The image response, the value of `success` should be checked before using the value of `message`
        """

        # Create the request body
        request_body = open_ai_models.ImageRequest(prompt=prompt, style=style)

        # Open a session
        async with aiohttp.ClientSession(
            headers=self._headers, base_url=constants.BASE_URL
        ) as session:
            # Send the request
            async with session.post(
                constants.IMAGE_URL, json=request_body.model_dump()
            ) as response:
                # Check the status code
                if response.status == 200:
                    # Parse the response body
                    response_body = open_ai_models.ImageResponse.model_validate_json(
                        await response.text()
                    )

                    # Create the response
                    response = SimpleOpenaiResponse(True, response_body.data[0].url)
                else:
                    # Parse the error response body
                    response_body = open_ai_models.ErrorResponse.model_validate_json(
                        await response.text()
                    )

                    # Create the response
                    response = SimpleOpenaiResponse(False, response_body.error.message)

                # Return the response
                return response
